"""Module with utils for training process."""

import random

import numpy as np
import torch


def seed_everything(seed: int) -> None:
    """Set all kinds of seeds.

    Args:
        seed: seed to be set
    """
    random.seed(seed)
    np.random.seed(seed)
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    torch.backends.cudnn.deterministic = True



"""Module with Pandas dataset implementation."""

from typing import Any

import numpy as np
import pandas as pd
from torch.utils.data import Dataset


class PandasDataset(Dataset):
    """Pandas dataset implementation."""
    def __init__(
        self,
        dataframe: pd.DataFrame,
        return_dicts: bool = False,
        target_col: str | None = None,
        mask_value: int | None = None,
        masking_proba: float | None = None,
        columns_to_mask: list[str] | None = None,
    ):
        """Instantiate dataset.

        Args:
            dataframe: input dataframe to wrap as toch dataset
            return_dicts: flag to return dict or tensor of values
            target_col: target column name
            mask_value: value to mask category columns
            masking_proba: probabilty to mask category values mask value
            columns_to_mask: columns to apply masking
        """
        self.dataframe = dataframe
        self.return_dicts = return_dicts
        self.target_col = target_col
        self.mask_value = mask_value
        self.masking_proba = masking_proba
        self.columns_to_mask = columns_to_mask or []

    def __len__(self) -> int:
        """Get dataset length.

        Returns:
            Dataset length
        """
        return len(self.dataframe)

    def  __getitem__(self, index: int) -> dict[str, Any] | Any:
        """Get dataset item by index.

        Args:
            index: index to take

        Returns:
            Dataset element under specified index
        """
        row = self.dataframe.iloc[index]

        # apply masking
        if self.mask_value is not None and self.masking_proba is not None:
            for col in self.columns_to_mask:
                if np.random.uniform() < self.masking_proba:
                    # to avoid SettingWithCopyWarning
                    row = row.where(row.index != col, self.mask_value)

        if self.return_dicts:
            return {k: v if not isinstance(v, list | tuple) else np.array(v) for k, v in row.to_dict().items()}

        if not self.target_col:
            return row.values

        return row.drop(self.target_col).values, row[self.target_col]



"""Module with utils for importing packages."""

import importlib
from typing import Any


def import_module_by_path(import_path: str) -> Any:
    """Import any object by its path.

    Args:
        import_path: path to object

    Returns:
        Imported object
    """
    module_name, object_name = import_path.rsplit(".", maxsplit=1)
    return getattr(importlib.import_module(module_name), object_name)



"""Module with logging setup implementation."""

import logging
import logging.config

import yaml

_CONF = """version: 1
disable_existing_loggers: False
formatters:
  default:
    format: '[%(asctime)s]{%(filename)s:%(lineno)s} - %(levelname)s - %(message)s'
handlers:
  default:
    formatter: default
    class: logging.StreamHandler
    stream: ext://sys.stderr
root:
  level: INFO
  handlers:
    - default
  propagate: no
"""


def get_logger(name: str) -> logging.Logger:
    """Get logger with given name.

    Args:
        name: logger name
    """
    logger = logging.getLogger(name)
    logging.config.dictConfig(yaml.safe_load(_CONF))
    return logger



"""Module FinalNet blocks implementation."""

from collections.abc import Callable, Sequence
from typing import Literal

import torch
from torch import nn


class FieldGate(nn.Module):
    """FieldGate block implementation."""
    def __init__(self, num_fields: int):
        """FieldGate block instantiation.

        Args:
            num_fields: number of fields to project
        """
        super().__init__()
        self.proj_field = nn.Linear(num_fields, num_fields)

    def forward(self, feature_emb: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Args:
            feature_emb: feature embeddings tensor

        Returns:
            Tensor after FieldGate applying
        """
        gates = self.proj_field(feature_emb.transpose(1, 2)).transpose(1, 2)
        out = torch.cat([feature_emb, feature_emb * gates], dim=1)
        return out


class FinalBlock(nn.Module):
    """FinalNet basic block implementation."""
    def __init__(
        self,
        input_dim: int,
        hidden_dims: Sequence[int],
        hidden_activations: Sequence[Callable] | Callable | None = None,
        dropout_rates: Sequence[float] | float | None = None,
        use_batch_norm: bool = True,
        add_bias: bool = True,
        residual_type: Literal["sum", "concat"] = "sum",
    ):
        """FinalNet basic block instantiation.

        Args:
            input_dim: number of input dimensions
            hidden_dims: hidden layers dimensions
            hidden_activations: hidden layers activations
            dropout_rates: hidden layers dropout rates
            use_batch_norm: use batch norm for hidden layers
            add_bias: add bias to FactorizedInteraction layer
            residual_type: type of residual for FactorizedInteraction, one of `sum` and `concat`
        """
        super().__init__()
        dropout_rates = dropout_rates or 0
        if not isinstance(dropout_rates, list):
            dropout_rates = [dropout_rates] * len(hidden_dims)

        hidden_activations = hidden_activations or nn.ReLU
        if not isinstance(hidden_activations, list):
            hidden_activations = [hidden_activations] * len(hidden_dims)

        hidden_dims = [input_dim, *hidden_dims]
        layers = []
        for idx, (in_dim, out_dim) in enumerate(zip(hidden_dims[:-1], hidden_dims[1:], strict=False)):
            layers.append(
                FactorizedInteraction(
                    input_dim=in_dim,
                    output_dim=out_dim,
                    bias=add_bias,
                    residual_type=residual_type,
                    activation_fn=None,
                )
            )
            if use_batch_norm:
                layers.append(nn.BatchNorm1d(out_dim))
            layers.append(hidden_activations[idx]())
            if dropout_rates[idx] > 0:
                layers.append(nn.Dropout(dropout_rates[idx]))

        self.net = nn.Sequential(*layers)

    def forward(self, input_: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Args:
            input_: input tensor

        Returns:
            Tensor after FinalBlock applying
        """
        output = self.net(input_)
        return output


class FactorizedInteraction(nn.Module):
    """Factorized interaction layer implementaion."""
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        bias: bool = True,
        residual_type: Literal["sum", "concat"] = "sum",
        activation_fn: Callable | None = None,
    ):
        """Factorized interaction layer instantiation.

        A replacement of nn.Linear to enhance multiplicative feature interactions.
        `residual_type="concat"` uses the same number of parameters as nn.Linear
        `residual_type="sum"` doubles the number of parameters.

        Args:
            input_dim: number of input dimensions
            output_dim: number of output dimensions
            bias: add bias to linear layer
            residual_type: type of residual, one of `sum` and `concat`
            activation_fn: activation function to use
        """
        super().__init__()
        if residual_type not in ["sum", "concat"]:
            raise RuntimeError(f"Unknown type of residual `{residual_type}`, should be on of [`sum`, `concat`]")

        if residual_type == "concat" and output_dim % 2:
            raise RuntimeError(f"`output_dim` should be divisible by 2, got {output_dim}")

        self.residual_type = residual_type
        if residual_type == "sum":
            output_dim = output_dim * 2
        self.linear = nn.Linear(input_dim, output_dim, bias=bias)

        activation_fn = activation_fn or nn.Identity
        self.activation = activation_fn()

    def forward(self, input_: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Args:
            input_: input tensor

        Returns:
            Tensor after FactorizedInteraction applying
        """
        linear_out = self.linear(input_)
        chunk_1, chunk_2 = torch.chunk(linear_out, chunks=2, dim=-1)

        chunk_2 = self.activation(chunk_2)
        if self.residual_type == "concat":
            return torch.cat([chunk_1, chunk_1 * chunk_2], dim=-1)

        return chunk_1 + chunk_2 * chunk_1



"""Module with basic trainer config implementation."""

from __future__ import annotations

import pathlib
from dataclasses import field
from typing import Literal

from pydantic.dataclasses import dataclass

from rankfx.common.base.config.base_config import BaseConfig


@dataclass
class TrainingConfig(BaseConfig):
    """Basic trainer config."""
    # data params
    batch_size: int = field(default=64)
    num_workers: int = field(default=12)
    return_dict_batches: bool = field(default=False)
    drop_last_batch: bool = field(default=False)

    # training params
    num_epochs: int = field(default=10)

    grad_clip_threshold: float = field(default=10.)

    seed: int = field(default=42)
    device: str = field(default="cpu")

    artifacts_path: pathlib.Path | str = field(default="./artifacts")
    checkpoint_filename: str = field(default="checkpoint.pth")
    best_model_filename: str = field(default="best_model.pth")

    # validation params
    validate_every_n_epochs: int = field(default=1)
    patience: int = field(default=5)
    eval_metric_name: str | None = field(default=None)
    eval_mode: Literal["min", "max"] = field(default="max")

    def checkpoint_path(self, **kwargs) -> pathlib.Path:
        """Get path to save checkpoint.

        Args:
            kwargs: will be used in checkpoint filename

        Returns:
            Checkpoint save path
        """
        checkpoint_filename = self.checkpoint_filename
        if kwargs:
            params_str = "_".join([f"{key}={val}" for key, val in kwargs.items()])
            checkpoint_filename = f"{params_str}_{checkpoint_filename}"
        return pathlib.Path(self.artifacts_path).resolve() / checkpoint_filename

    @property
    def best_model_path(self) -> pathlib.Path:
        """Get path to save best model.

        Returns:
            Best model save path
        """
        return pathlib.Path(self.artifacts_path).resolve() / self.best_model_filename



"""Module with available features types listing."""

import enum


@enum.unique
class FeatureType(enum.Enum):
    """Available feature types."""
    NUMERICAL = "numerical"
    CATEGORICAL = "categorical"
    NUMERICAL_SEQUENCE = "numerical_sequence"
    CATEGORICAL_SEQUENCE = "categorical_sequence"



"""Basic MLP net module implementation."""

from collections.abc import Callable, Sequence

import torch
from torch import nn


class MLPBlock(nn.Module):
    """MLP block implementation."""
    def __init__(
        self,
        in_features: int,
        out_features: int | None = None,
        hidden_dims: Sequence[int] | None = None,
        activation_fn: Callable = nn.ReLU,
        dropout: float = 0.,
        use_batch_norm: bool = True,
        add_bias: bool = True,
    ):
        """Instantiate MLPBlock module.

        Args:
            in_features: number of input features
            out_features: number of output features
            hidden_dims: hidden net dimensions
            activation_fn: callable to get activation for hidden layers
            dropout: dropout rate
            use_batch_norm: use batch norm for hidden blocks
            add_bias: add bias for linear layers
        """
        super().__init__()

        if out_features is None and hidden_dims is None:
            raise RuntimeError("both `out_features` and `hidden_dims` are None, set at least one of them")

        hidden_dims = [in_features, *(hidden_dims or [])]

        layers = []
        for in_dim, out_dim in zip(hidden_dims[:-1], hidden_dims[1:], strict=False):
            layers.append(nn.Linear(in_dim, out_dim, bias=add_bias))
            if use_batch_norm:
                layers.append(nn.BatchNorm1d(out_dim))
            layers.append(activation_fn())
            if dropout > 0:
                layers.append(nn.Dropout(p=dropout))

        if out_features is not None:
            layers.append(nn.Linear(hidden_dims[-1], out_features, bias=add_bias))
        self.mlp = nn.Sequential(*layers)

    def forward(self, input_: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Args:
            input_: input tensor

        Returns:
            Tensor after MLPBlock applying
        """
        return self.mlp(input_)



"""Package with DCNv2 modules implementation."""

from .crossnet import CrossNetMix, CrossNetV2
from .embedding import EmbeddingLayer
from .finalnet import FactorizedInteraction, FieldGate, FinalBlock
from .mlp import MLPBlock

__all__ = [
    "CrossNetMix",
    "CrossNetV2",
    "EmbeddingLayer",
    "FactorizedInteraction",
    "FieldGate",
    "FinalBlock",
    "MLPBlock",
]



"""Package with base entities implementation."""

from .config import BaseConfig, TrainingConfig
from .model import ModelPhase, NNPandasModel

__all__ = ["BaseConfig", "ModelPhase", "NNPandasModel", "TrainingConfig"]



"""Module with Embedding layer implementation."""

import torch
from torch import nn

from rankfx.common.features.config import FeaturesConfig
from rankfx.common.features.types import FeatureType


class EmbeddingLayer(nn.Module):
    """Embedding Layer implementation."""
    def __init__(
        self,
        features_config: FeaturesConfig,
        is_dict_input: bool = True,
        flatten_emb: bool = True,
        proj_output_features: bool = False,
    ):
        """Instantiate EmbeddingLayer module.

        Features are taken in the order, specified in features_config.features
        If is_dict_input is False, feature will be taken as slices from input tensor

        Args:
            features_config: features config instance
            is_dict_input: is input looks like {feature_name: feature_tensor} or just simple tensor
            flatten_emb: do flatten embeddings output into single vector, or try to stack by features
            proj_output_features: apply projection to output concatted embeddings
        """
        super().__init__()
        self.config = features_config
        self.is_dict_input = is_dict_input
        self.flatten_emb = flatten_emb
        self.proj_output_features = proj_output_features

        modules = {}
        for feature in features_config.features:
            if feature.needs_embed:
                if feature.feature_type == FeatureType.CATEGORICAL:
                    modules[feature.name] = nn.Embedding(
                        feature.embedding_vocab_size,
                        feature.embedding_size,
                        padding_idx=feature.embedding_padding_idx,
                    )
                    continue

                if feature.feature_type == FeatureType.CATEGORICAL_SEQUENCE:
                    modules[feature.name] = nn.EmbeddingBag(
                        feature.embedding_vocab_size,
                        feature.embedding_size,
                        padding_idx=feature.embedding_padding_idx,
                    )
                    continue

                modules[feature.name] = nn.Linear(feature.feature_size, feature.embedding_size, bias=False)

        self.output_proj = nn.Linear(
            in_features=features_config.num_final_features,
            out_features=features_config.num_final_features,
            bias=False,
        ) if proj_output_features else nn.Identity()

        self.embedding_modules = nn.ModuleDict(modules)
        self.dummy_fn = nn.Identity()

    def forward(self, input_: dict[str, torch.Tensor] | torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Args:
            input_: input tensor [{feature_name: feature_tensor} or tensor (batch_size, concat_features_len)]

        Returns:
            Tensor after EmbeddingLayer applying
        """
        if not self.is_dict_input and self.config.num_initial_features != input_.size(-1):
            raise RuntimeError(
                f"Number of features in config ({self.config.num_initial_features})"
                f"and in input tensor ({input_.size(-1)}) are not equal"
            )

        outputs = []
        current_idx = 0
        for feature in self.config.features:
            feature_value = (
                input_[feature.name]
                if self.is_dict_input
                else input_[:, current_idx:current_idx + feature.feature_size]
            )
            if feature.feature_type in [FeatureType.CATEGORICAL, FeatureType.CATEGORICAL_SEQUENCE]:
                feature_value = feature_value.long()
                if feature.needs_embed:
                    feature_value = feature_value.squeeze()
                elif feature.feature_size == 1:
                    feature_value = feature_value.unsqueeze(1)

            if feature.feature_type in [FeatureType.NUMERICAL, FeatureType.NUMERICAL_SEQUENCE]:
                feature_value = feature_value.type(torch.float32)
                if feature.feature_size == 1:
                    feature_value = feature_value.unsqueeze(1)

            if feature.name in self.embedding_modules:
                output = self.embedding_modules[feature.name](feature_value)
            else:
                output = self.dummy_fn(feature_value)

            current_idx += feature.feature_size
            outputs.append(output)

        if self.flatten_emb:
            # (batch_size, concat_features_len)
            return self.output_proj(torch.cat(outputs, dim=-1))

        # (batch_size, feature_num, feature_len) - features should be the same size
        return torch.stack(outputs, dim=1)



"""Package with features preprocessing/configuration and logic."""

from .config import Feature, FeaturesConfig
from .datasets import PandasDataset
from .types import FeatureType

__all__ = ["Feature", "FeatureType", "FeaturesConfig", PandasDataset]



"""Module with base model implementation."""

import enum
import gc
import os
import pathlib
from abc import ABC, abstractmethod
from collections.abc import Sequence
from typing import Any, Literal

import numpy as np
import pandas as pd
import torch
from pandas.api.types import is_list_like, is_numeric_dtype
from torch import nn
from torch.optim import Optimizer
from torch.optim.lr_scheduler import LRScheduler, ReduceLROnPlateau
from torch.utils.data import DataLoader
from tqdm.auto import tqdm

from rankfx.common.features.config import Feature, FeaturesConfig, FeatureType
from rankfx.common.features.datasets import PandasDataset
from rankfx.common.modules import EmbeddingLayer
from rankfx.common.utils.import_utils import import_module_by_path
from rankfx.common.utils.logging_utils import get_logger
from rankfx.common.utils.training_utils import seed_everything

LOGGER = get_logger(__name__)


@enum.unique
class ModelPhase(enum.Enum):
    """Possible model usage phases."""
    TRAIN = "Train"
    VALIDATION = "Validation"
    TEST = "Test"
    INFERENCE = "Inference"


class NNPandasModel(ABC, nn.Module):
    """Base model class."""
    def __init__(
        self,
        infer_feature_config: bool = True,
        oov_idx: int = 0,
    ):
        """Instantiate model class.

        Args:
            infer_feature_config: infer features config from input pandas dataframe
            oov_idx: index to use as OOV for all categorical features embeddings
        """
        super().__init__()
        self.infer_feature_config = infer_feature_config
        self.oov_idx = oov_idx

        # categorical features mappings
        self._feature_mappings: dict[str, dict[Any, int]] | None = None

    @abstractmethod
    def forward(self, input_: dict[str, torch.Tensor] | torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Args:
            input_: input tensor [{feature_name: feature_tensor} or tensor (batch_size, concat_features_len)]

        Returns:
            Tensor after model applying
        """
        raise NotImplementedError

    @abstractmethod
    def train_step(self, batch: dict[str, torch.Tensor] | torch.Tensor) -> dict[str, Any]:
        """Model training step.

        Output MUST have `loss` key

        Args:
            batch: batch of data

        Returns:
            Loss/metrics after training step
        """
        raise NotImplementedError

    @abstractmethod
    def val_step(self, batch: dict[str, torch.Tensor] | torch.Tensor) -> dict[str, Any]:
        """Model validation step.

        Args:
            batch: batch of data

        Returns:
            Loss/metrics after validation step
        """
        raise NotImplementedError

    @abstractmethod
    def test_step(self, batch: dict[str, torch.Tensor] | torch.Tensor) -> dict[str, Any]:
        """Model testing step.

        Args:
            batch: batch of data

        Returns:
            Loss/metrics after testing step
        """
        raise NotImplementedError

    @abstractmethod
    def inference_step(self, batch: dict[str, torch.Tensor] | torch.Tensor) -> Any:
        """Model inference step.

        Args:
            batch: batch of data

        Returns:
            Model inference output
        """
        raise NotImplementedError

    @abstractmethod
    def _init_modules(self, features_config: FeaturesConfig) -> None:
        """Instantiate model modules based on features config.

        Args:
            features_config: features config
        """
        raise NotImplementedError

    def _init_optimizers(
        self,
        optimizer_cls: str | None = None,
        optimizer_params: dict[str, Any] | None = None,
        scheduler_cls: str | None = None,
        scheduler_params: dict[str, Any] | None = None,
    ) -> tuple[Optimizer | None, LRScheduler | None]:
        """Instatiate optimizer and scheduler objects if provided.

        Args:
            optimizer_cls: optimizer import path
            optimizer_params: optimizer parameters
            scheduler_cls: scheduler import path
            scheduler_params: scheduler parameters

        Returns:
            Tuple of optimizer and scheduler
        """
        optimizer = None
        if optimizer_cls:
            optimizer_cls = import_module_by_path(optimizer_cls)
            optimizer = optimizer_cls(self.parameters(), **(optimizer_params or {}))

        scheduler = None
        if scheduler_cls and optimizer is not None:
            scheduler_cls = import_module_by_path(scheduler_cls)
            scheduler = scheduler_cls(optimizer, **(scheduler_params or {}))

        return optimizer, scheduler

    def _infer_features_config_from_dataframe(
        self,
        data: pd.DataFrame,
        default_embedding_size: int = 10,
        custom_embedding_sizes: dict[str, int] | None = None,
        embedded_features: Sequence[str] | None = None,
    ) -> FeaturesConfig:
        """Create feature config from pandas dataframe.

        Args:
            data: pandas dataframe to infrence features config
            default_embedding_size: default features embedding size
            custom_embedding_sizes: custom embeddings mapping {feature: feature: embedding_size}
            embedded_features: features to embed

        Returns:
            Created features config
        """
        if not len(data):
            raise RuntimeError("Got empty dataframe for inferencing config")

        custom_embedding_sizes = custom_embedding_sizes or {}
        features = []
        for col in data.columns:
            num_embed_params = {}
            if col in embedded_features:
                num_embed_params = {
                    "needs_embed": True,
                    "embedding_size": custom_embedding_sizes.get(col, default_embedding_size),
                }

            if is_numeric_dtype(data[col].dtype):
                features.append(Feature(name=col, feature_type=FeatureType.NUMERICAL, **num_embed_params))
                continue

            col_value = data.at[0, col]
            if isinstance(data[col].dtype, pd.CategoricalDtype) or isinstance(col_value, str):
                # saving one index for oov values
                vocab_size = data[col].nunique() + 1
                features.append(
                    Feature(
                        name=col,
                        feature_type=FeatureType.CATEGORICAL,
                        needs_embed=True,
                        embedding_size=custom_embedding_sizes.get(col, default_embedding_size),
                        embedding_vocab_size=vocab_size,
                    )
                )
                continue

            if not is_list_like(col_value):
                raise RuntimeError(
                    f"Feature `{col}` is not category, not numerical and not list-like, got {col_value}"
                )

            feature_size = data[col].apply(len).unique()
            if len(feature_size) > 1:
                raise RuntimeError(f"Got multiple sequence lengths for feature `{col}`")
            feature_size = feature_size[0]

            if is_numeric_dtype(np.concat(data[col].values)):
                features.append(
                    Feature(
                        name=col,
                        feature_type=FeatureType.NUMERICAL_SEQUENCE,
                        feature_size=feature_size,
                        **num_embed_params,
                    )
                )
                continue

            # saving one index for oov values
            vocab_size = data[col].explode().nunique() + 1
            features.append(
                Feature(
                    name=col,
                    feature_type=FeatureType.CATEGORICAL_SEQUENCE,
                    feature_size=feature_size,
                    needs_embed=True,
                    embedding_size=custom_embedding_sizes.get(col, default_embedding_size),
                    embedding_vocab_size=vocab_size,
                )
            )

        return FeaturesConfig(features=features)

    def _build_features_mappings(self, data: pd.DataFrame) -> dict[str, dict[Any, int]]:
        """Build categorical features mappings.

        Args:
            data: dataframe to build mappings

        Returns:
            Built mappings
        """
        if not len(data):
            raise RuntimeError("Got empty dataframe for building categorical mappings")

        features_mappings = {}
        for col in data.columns:
            col_value = data.at[0, col]
            feature_values = data[col].explode() if is_list_like(col_value) else data[col]

            if is_numeric_dtype(feature_values.dtype):
                continue

            if not (isinstance(feature_values.dtype, pd.CategoricalDtype) or isinstance(col_value, str)):
                raise RuntimeError(
                    f"Feature `{col}` is not category, not numerical and not string, got {col_value}"
                )

            LOGGER.info("Building mapping for feature %s", col)
            uniq_values = feature_values.unique().tolist()
            indexes = list(range(len(uniq_values) + 1))

            if self.oov_idx not in indexes:
                raise RuntimeError("OOV idx is not in interval for feature %s", col)

            indexes.pop(self.oov_idx)
            features_mappings[col] = dict(zip(uniq_values, indexes, strict=False))

        return features_mappings

    def _encode_categorical_features(self, features: pd.DataFrame) -> pd.DataFrame:
        """Encode categorical features using vocabulary.

        Args:
            features: dataframe with input features
        Returns:
            Encoded dataframe
        """
        if self._feature_mappings is None:
            raise RuntimeError("Trying to encode features, but features mapping are not built")

        for col, mapping in self._feature_mappings.items():
            if col in features:
                LOGGER.info("Encoding feature %s", col)
                # getting rid of possible category dtypes
                features[col] = features[col].map(mapping).astype(float)
                features[col] = features[col].fillna(self.oov_idx).astype(int)

        return features

    def _get_dataloader_from_dataframes(
        self,
        features: pd.DataFrame,
        target: pd.Series | np.ndarray | None = None,
        shuffle: bool = False,
        batch_size: int = 64,
        num_workers: int = 12,
        drop_last_batch: bool = False,
        masking_proba: float | None = None,
    ) -> DataLoader:
        """Create dataloaders from input pandas dataframes.

        Args:
            features: dataframe with input features
            target: dataframe with train target
            shuffle: shuffle elements flag
            batch_size: size of dataloader batches
            num_workers: number of dataloader workers
            drop_last_batch: drop last incomplete batch
            masking_proba: probabilty to mask category values with oov
        Returns:
            Created dataloader
        """
        features = features.copy(deep=True)
        columns_to_mask = None

        if self._feature_mappings is not None:
            features = self._encode_categorical_features(features=features)
            columns_to_mask = list(self._feature_mappings.keys())

        target_col = None
        if target is not None:
            features["target"] = target
            target_col = "target"

        dataset = PandasDataset(
            features,
            return_dicts=True,
            target_col=target_col,
            mask_value=self.oov_idx,
            masking_proba=masking_proba,
            columns_to_mask=columns_to_mask,
        )

        return DataLoader(
            dataset,
            batch_size=batch_size,
            shuffle=shuffle,
            num_workers=num_workers,
            pin_memory=True,
            drop_last=drop_last_batch,
        )

    def _move_batch_to_device(
        self,
        batch: dict[str, torch.Tensor] | torch.Tensor,
        device: torch.device,
    ) -> dict[str, torch.Tensor] | torch.Tensor:
        """Move whole batch to device.

        Args:
            batch: input batch of data
            device: device to put batch on

        Returns:
            Moved batch
        """
        if isinstance(batch, torch.Tensor):
            return batch.to(device)

        if isinstance(batch, dict):
            for key in batch:
                batch[key] = batch[key].to(device)
        else:
            for idx in range(len(batch)):
                batch[idx] = batch[idx].to(device)
        return batch

    def _clear_resources(self) -> None:
        """Clear temporary resources."""
        gc.collect()
        torch.cuda.empty_cache()

    def _check_metrics_improvement(
        self,
        new_value: float,
        prev_value: float | None,
        eval_mode: Literal["min", "max"] = "max",
    ) -> bool:
        """Check if metrics are improving based on eval mode.

        Args:
            new_value: new metric value
            prev_value: previous metric value
            eval_mode: eval rule for metrics

        Returns:
            True if metric has improved
        """
        if prev_value is None:
            return True

        if eval_mode == "min":
            return new_value <= prev_value

        if eval_mode == "max":
            return new_value >= prev_value

        raise RuntimeError(f"Got unknown eval_mode {eval_mode}, should be on of (`min`, `max`)")

    def _calculate_l2_term(self, l2_net_reg: float = 0., l2_embedding_reg: float = 0.,) -> float | torch.Tensor:
        """Calculate L2 regularization term.

        Args:
            l2_net_reg: regularizer term for net parameters
            l2_embedding_reg: regularizer term for embeddings parameters

        Returns:
            L2 regularization term
        """
        if not (l2_embedding_reg or l2_net_reg):
            return 0.

        reg_term, emb_params = 0, set()
        for module_name, module in self.named_modules():
            if not isinstance(module, EmbeddingLayer):
                continue

            for p_name, param in module.named_parameters():
                if param.requires_grad:
                    emb_params.add(".".join([module_name, p_name]))
                    reg_term += 0.5 * l2_embedding_reg * torch.norm(param, 2) ** 2

        for p_name, param in self.named_parameters():
            if param.requires_grad:
                if p_name not in emb_params:
                    reg_term += 0.5 * l2_net_reg * torch.norm(param, 2) ** 2

        return reg_term

    def _run_model_step(
        self,
        batch: dict[str, torch.Tensor] | torch.Tensor,
        phase: ModelPhase,
    ) -> Any:
        """Run model step depending on current phase.

        Args:
            batch: batch of data
            phase: model phase

        Returns:
            Loss/metrics after step
        """
        if phase == ModelPhase.TRAIN:
            return self.train_step(batch)

        if phase == ModelPhase.VALIDATION:
            return self.val_step(batch)

        if phase == ModelPhase.TEST:
            return self.test_step(batch)

        return self.inference_step(batch)

    def _run_inference_epoch(
        self,
        dataloader: DataLoader,
        device: torch.device | str,
    ) -> torch.Tensor:
        """Run single inference epoch.

        Args:
            dataloader: torch dataloader
            device: working device

        Returns:
            Output tensors after model inference
        """
        self.eval()

        batch_results = []

        progress_bar = tqdm(dataloader, desc="Inference epoch")
        for batch in progress_bar:
            batch = self._move_batch_to_device(batch, device=device)
            with torch.inference_mode():
                step_output = self._run_model_step(batch, phase=ModelPhase.INFERENCE)
            batch_results.append(step_output)
        return torch.cat(batch_results, dim=0)

    def _run_epoch(
        self,
        dataloader: DataLoader,
        epoch_num: int,
        optimizer: Optimizer | None = None,
        phase: ModelPhase = ModelPhase.TRAIN,
        scheduler: LRScheduler | None = None,
        device: torch.device | str = "cpu",
        grad_clip_threshold: float | None = None,
        l2_net_reg: float = 0.,
        l2_embedding_reg: float = 0.,
    ) -> dict[str, Any]:
        """Run single epoch.

        Args:
            dataloader: torch dataloader
            epoch_num: number of epoch
            optimizer: model optimizer
            device: working device
            phase: model phase
            scheduler: learning rate scheduler
            grad_clip_threshold: value to clip gradients
            l2_net_reg: regularizer term for net parameters
            l2_embedding_reg: regularizer term for embeddings parameters

        Returns:
            Average metrics after training epoch
        """
        if phase == ModelPhase.TRAIN:
            self.train()
            optimizer.zero_grad()
        else:
            self.eval()

        steps_per_epoch = 0
        epoch_metrics = {}

        progress_bar = tqdm(dataloader, desc=f"{phase.value} epoch #{epoch_num}")
        for batch in progress_bar:
            batch = self._move_batch_to_device(batch, device=device)
            steps_per_epoch += 1

            if phase == ModelPhase.TRAIN:
                optimizer.zero_grad()
                step_output = self._run_model_step(batch, phase=phase)
                loss = (
                    step_output["loss"]
                    + self._calculate_l2_term(l2_net_reg=l2_net_reg, l2_embedding_reg=l2_embedding_reg)
                )
                loss.backward()

                if grad_clip_threshold:
                    torch.nn.utils.clip_grad_norm_(self.parameters(), grad_clip_threshold)

                optimizer.step()
                if scheduler is not None and not isinstance(scheduler, ReduceLROnPlateau):
                    scheduler.step()
            else:
                with torch.inference_mode():
                    step_output = self._run_model_step(batch, phase=phase)

            for key, value in step_output.items():
                epoch_metrics[key] = epoch_metrics.get(key, 0) + value

            epoch_metrics_str = ", ".join(
                [f"{key}: {value / steps_per_epoch:.5f}" for key, value in epoch_metrics.items()]
            )
            # update progress bar
            progress_bar.set_postfix(metrics=f"[{epoch_metrics_str}]")

        epoch_metrics = {key: value / steps_per_epoch for key, value in epoch_metrics.items()}
        epoch_metrics_str = ", ".join([f"{key}: {value:.5f}" for key, value in epoch_metrics.items()])
        LOGGER.info("Finished %s Epoch #%s, average metrics - [%s]", phase.value, epoch_num, epoch_metrics_str)

        self._clear_resources()
        return epoch_metrics

    def get_feature_mapping(self) -> dict[str, dict[Any, int]] | None:
        """Get categorical features mappings.

        Returns:
            Categorical features mappings
        """
        return self._feature_mappings

    def set_feature_mapping(self, mappings: dict[str, dict[Any, int]] | None) -> None:
        """Set categorical features mappings.

        Args:
            mappings: categorical features mappings to set
        """
        self._feature_mappings = mappings

    def fit(
        self,
        features: pd.DataFrame,
        target: pd.Series | np.ndarray = None,
        val_features: pd.DataFrame | None = None,
        val_target: pd.Series | np.ndarray = None,
        optimizer_cls: str | None = None,
        optimizer_params: dict[str, Any] | None = None,
        scheduler_cls: str | None = None,
        scheduler_params: dict[str, Any] | None = None,
        num_epochs: int = 10,
        seed: int | None = None,
        artifacts_path: str | None = None,
        device: str | torch.device = "cpu",
        batch_size: int = 64,
        num_workers: int = 12,
        drop_last_batch: bool = False,
        grad_clip_threshold: float | None = None,
        validate_every_n_epochs: int = 1,
        eval_metric_name: str | None = None,
        eval_mode: Literal["min", "max"] = "max",
        patience: int | None = None,
        default_embedding_size: int = 10,
        custom_embedding_sizes: dict[str, int] | None = None,
        embedded_features: Sequence[str] | None = None,
        features_mappings: dict[str, dict[Any, int]] | None = None,
        oov_masking_proba: float | None = None,
        l2_net_reg: float = 0.,
        l2_embedding_reg: float = 0.,
    ) -> tuple[dict[str, Any], dict[str, Any] | None]:
        """Train model using pandas dataframes.

        Args:
            features: features to train
            target: target to train
            val_features: features for validation, if None dont validate
            val_target: target for validation
            optimizer_cls: optimizer import path
            optimizer_params: optimizer parameters
            scheduler_cls: scheduler import path
            scheduler_params: scheduler parameters
            num_epochs: number of training epochs
            seed: random seed
            artifacts_path: path to save artifacts
            device: working device
            batch_size: size of dataloader batches
            num_workers: number of dataloader workers
            drop_last_batch: drop last incomplete batch
            grad_clip_threshold: value to clip gradients
            validate_every_n_epochs: evaulation frequency
            eval_metric_name: metric name to evaluate
            eval_mode: eval rule for metrics
            patience: number of epochs to wait for metrics improvement, after that stop training
            default_embedding_size: default features embedding size
            custom_embedding_sizes: custom embeddings mapping {feature: feature: embedding_size}
            embedded_features: features to embed
            features_mappings: categorical features mappings to use, if None will be built from train dataframe
            oov_masking_proba: probability to mask categorical features with oov_idx
            l2_net_reg: regularizer term for net parameters
            l2_embedding_reg: regularizer term for embeddings parameters

        Returns:
            Output metrics from train and validation steps
        """
        if self.infer_feature_config:
            features_config = self._infer_features_config_from_dataframe(
                data=features,
                default_embedding_size=default_embedding_size,
                custom_embedding_sizes=custom_embedding_sizes,
                embedded_features=embedded_features or [],
            )
            self._init_modules(features_config=features_config)
        LOGGER.info("Used features config: %s", features_config)

        optimizer, scheduler = self._init_optimizers(
            optimizer_cls=optimizer_cls,
            optimizer_params=optimizer_params,
            scheduler_cls=scheduler_cls,
            scheduler_params=scheduler_params,
        )

        eval_metric_name = eval_metric_name or "loss"
        if seed:
            seed_everything(seed)

        best_model_path = None
        if artifacts_path:
            LOGGER.info("Artifacts path is %s", pathlib.Path(artifacts_path).resolve())
            if not os.path.isdir(artifacts_path):
                LOGGER.info("Creating artifacts path")
                os.makedirs(artifacts_path)

            best_model_path = str(pathlib.Path(artifacts_path) / "best_model.pt")
            LOGGER.info("Best model path is %s", best_model_path)

        self = self.to(device)
        optimizer = optimizer or torch.optim.Adam(self.parameters())

        LOGGER.info("Building features mappings")
        self._feature_mappings = features_mappings or self._build_features_mappings(data=features)

        LOGGER.info("Building train dataloader")
        train_dataloader = self._get_dataloader_from_dataframes(
            features=features,
            target=target,
            shuffle=True,
            batch_size=batch_size,
            num_workers=num_workers,
            drop_last_batch=drop_last_batch,
            masking_proba=oov_masking_proba,
        )

        val_dataloader = None
        if val_features is not None:
            LOGGER.info("Building validation dataloader")
            val_dataloader = self._get_dataloader_from_dataframes(
                features=val_features,
                target=val_target,
                shuffle=False,
                batch_size=batch_size,
                num_workers=num_workers,
                drop_last_batch=False,
            )

        LOGGER.info("Starting training process")
        train_output, val_output = None, None
        patience_epochs, best_eval_metric = 0, None

        for epoch in range(num_epochs):
            train_output = self._run_epoch(
                train_dataloader,
                epoch_num=epoch,
                phase=ModelPhase.TRAIN,
                optimizer=optimizer,
                scheduler=scheduler,
                device=device,
                grad_clip_threshold=grad_clip_threshold,
                l2_net_reg=l2_net_reg,
                l2_embedding_reg=l2_embedding_reg,
            )

            if (
                val_dataloader is not None
                and (not epoch % validate_every_n_epochs or epoch == num_epochs - 1)
            ):
                val_output = self._run_epoch(
                    val_dataloader,
                    epoch_num=epoch,
                    phase=ModelPhase.VALIDATION,
                    device=device,
                )

                current_eval_metric = val_output[eval_metric_name]
                if self._check_metrics_improvement(current_eval_metric, best_eval_metric, eval_mode=eval_mode):
                    patience_epochs = 0
                    best_eval_metric = current_eval_metric

                    if best_model_path:
                        torch.save(self.state_dict(), best_model_path)
                        LOGGER.info(
                            "Best model with %s = %s was saved to %s",
                            eval_metric_name,
                            best_eval_metric,
                            best_model_path,
                        )
                    continue

                if isinstance(scheduler, ReduceLROnPlateau):
                    scheduler.step(current_eval_metric)

                patience_epochs += 1
                if patience and patience_epochs > patience:
                    LOGGER.info("Metrics not increasing during %s epochs. Stop training", patience)
                    break

        LOGGER.info("Loading best model from %s", best_model_path)
        self.load_state_dict(torch.load(best_model_path, weights_only=True), strict=True)

        return train_output, val_output

    def test(
        self,
        features: pd.DataFrame,
        target: pd.Series | np.ndarray = None,
        device: str | torch.device = "cpu",
        batch_size: int = 64,
        num_workers: int = 12,
    ) -> dict[str, Any]:
        """Run prediction on pandas dataframe.

        Args:
            features: features for test
            target: target to for test
            device: working device
            batch_size: size of dataloader batches
            num_workers: number of dataloader workers

        Returns:
            Test metrics
        """
        self = self.to(device)
        LOGGER.info("Building test dataloader")
        dataloader = self._get_dataloader_from_dataframes(
            features=features,
            target=target,
            shuffle=False,
            batch_size=batch_size,
            num_workers=num_workers,
            drop_last_batch=False,
        )
        test_output = self._run_epoch(
            dataloader,
            epoch_num=-1,
            phase=ModelPhase.TEST,
            device=device,
        )
        return test_output

    def predict(
        self,
        features: pd.DataFrame,
        device: str | torch.device = "cpu",
        batch_size: int = 64,
        num_workers: int = 12,
    ) -> np.ndarray:
        """Run prediction on pandas dataframe.

        Args:
            features: features to train
            device: working device
            batch_size: size of dataloader batches
            num_workers: number of dataloader workers

        Returns:
            Predicted items
        """
        self = self.to(device)
        LOGGER.info("Building inference dataloader")
        dataloader = self._get_dataloader_from_dataframes(
            features=features,
            target=None,
            shuffle=False,
            batch_size=batch_size,
            num_workers=num_workers,
            drop_last_batch=False,
        )
        inference_output = self._run_inference_epoch(dataloader, device=device)
        return inference_output.detach().cpu().numpy()



"""Package with common utils for different models."""

from .import_utils import import_module_by_path
from .logging_utils import get_logger
from .training_utils import seed_everything

__all__ = ["get_logger", "import_module_by_path", "seed_everything"]



"""Module with CrossNet implementaion."""

import torch
from torch import nn


class CrossNetV2(nn.Module):
    """CrossNetv2 implementation."""
    def __init__(self, in_features: int, num_layers: int):
        """Instantiate CrossNetv2 module.

        Args:
            in_features: number of input features
            num_layers: numer of net layers
        """
        super().__init__()
        self.cross_layers = nn.ModuleList(
            nn.Linear(in_features, in_features) for _ in range(num_layers)
        )

    def forward(self, input_: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Args:
            input_: input tensor

        Returns:
            Tensor after CrossNet applying
        """
        output_i = input_
        for layer in self.cross_layers:
            output_i = input_ * layer(output_i) + output_i
        return output_i


class CrossNetMix(nn.Module):
    """CrossNetMix implementation.

    Improves CrossNetV2 by:
        1. Add MOE to learn feature interactions in different subspaces
        2. Add nonlinear transformations in low-dimensional space
    """
    def __init__(self, in_features: int, num_layers: int = 2, low_rank_dim: int = 32, num_experts: int = 4):
        """Instantiate CrossNetMix module.

        Args:
            in_features: number of input features
            num_layers: numer of net layers
            low_rank_dim: dimension of inner subspace
            num_experts: number of experts in MoE
        """
        super().__init__()
        self.num_layers = num_layers
        self.num_experts = num_experts

        # U: (in_features, low_rank)
        self.U_layers = nn.ParameterList(
            nn.Parameter(nn.init.xavier_normal_(torch.empty(num_experts, in_features, low_rank_dim)))
            for _ in range(self.num_layers)
        )
        # V: (in_features, low_rank)
        self.V_layers = nn.ParameterList(
            nn.Parameter(nn.init.xavier_normal_(torch.empty(num_experts, in_features, low_rank_dim)))
            for _ in range(self.num_layers)
        )
        # C: (low_rank, low_rank)
        self.C_layers = nn.ParameterList(
            nn.Parameter(nn.init.xavier_normal_(torch.empty(num_experts, low_rank_dim, low_rank_dim)))
            for _ in range(self.num_layers)
        )
        self.gating = nn.ModuleList(nn.Linear(in_features, 1, bias=False) for _ in range(self.num_experts))

        self.bias = nn.ParameterList(
            nn.Parameter(nn.init.zeros_(torch.empty(in_features, 1)))
            for _ in range(self.num_layers)
        )

    def forward(self, input_: torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Args:
            input_: input tensor

        Returns:
            Tensor after CrossNetMix applying
        """
        # input_: (batch_size, in_features, 1)
        initial_state = input_.unsqueeze(-1)
        interm_state = initial_state
        for layer_idx in range(self.num_layers):
            output_of_experts = []
            gating_score_of_experts = []
            for expert_idx in range(self.num_experts):
                # (1) G(x_l)
                # compute the gating score by x_l
                gating_score_of_experts.append(self.gating[expert_idx](interm_state.squeeze(-1)))

                # (2) E(x_l)
                # project the input x_l to $\mathbb{R}^{r}$
                # layer_output: (batch_size, low_rank_dim, 1)
                layer_output = torch.matmul(self.V_layers[layer_idx][expert_idx].T, interm_state)

                # nonlinear activation in low rank space
                layer_output = torch.tanh(layer_output)
                layer_output = torch.matmul(self.C_layers[layer_idx][expert_idx], layer_output)
                layer_output = torch.tanh(layer_output)

                # project back to $\mathbb{R}^{d}$
                # layer_output: (batch_size, in_features, 1)
                layer_output = torch.matmul(self.U_layers[layer_idx][expert_idx], layer_output)

                # Hadamard-product
                layer_output += self.bias[layer_idx]
                layer_output = initial_state * layer_output

                output_of_experts.append(layer_output.squeeze(-1))

            # (3) mixture of low-rank experts
            # output_of_experts: (batch_size, in_features, num_experts)
            output_of_experts = torch.stack(output_of_experts, -1)

            # gating_score_of_experts: (batch_size, num_experts, 1)
            gating_score_of_experts = torch.stack(gating_score_of_experts, -2)
            moe_out = torch.matmul(output_of_experts, gating_score_of_experts.softmax(-1))

            # interm_state: (batch_size, in_features, 1)
            interm_state = moe_out + interm_state

        # interm_state: (batch_size, in_features)
        interm_state = interm_state.squeeze(-1)
        return interm_state



"""Setup file for package."""

from setuptools import find_packages, setup


def readme():
  """Get README.md content."""
  with open("README.md") as f:
    return f.read()


setup(
  name="rankfx",
  version="1.0.0",
  author="saraevn",
  author_email="saraevnik0909@gmail.com",
  description="Python package for working with NN models",
  long_description=readme(),
  long_description_content_type="text/markdown",
  url="https://github.com/sn09/rankfx",
  packages=find_packages(exclude=["notebooks"]),
  install_requires=[
    "torch>=2.2.1",
    "pydantic>=2.10.6",
    "numpy>=2.2.4",
    "pandas>=2.2.3",
    "tqdm>=4.67.1",
    "scikit-learn>=1.6.1",
],
  classifiers=[
    "Programming Language :: Python :: 3.10",
    "License ::  MIT License",
    "Operating System :: OS Independent"
  ],
  keywords="ranking torch pandas",
  project_urls={
    "Documentation": "link"
  },
  python_requires=">=3.10"
)



"""Module with basic config implementation."""

from __future__ import annotations

import pathlib
from dataclasses import asdict
from typing import Any

import yaml
from pydantic.dataclasses import ConfigDict, dataclass


@dataclass
class BaseConfig:
    """Base config implementation."""
    model_config = ConfigDict(extra="ignore")

    def as_dict(self) -> dict[str, Any]:
        """Get config dict representation."""
        return asdict(self)

    @staticmethod
    def from_yaml(filepath: str | pathlib.Path) -> BaseConfig:
        """Load config from yaml file.

        Args:
            filepath: path to yaml file

        Returns:
            Config instance
        """
        with open(filepath) as fin:
            content = yaml.safe_load(fin)
        return BaseConfig(**content)

    def to_yaml(self) -> str:
        """Export config to yaml format.

        Returns:
            Config in yaml string format
        """
        return yaml.dump(self.as_dict())



"""Module with DCNv2 model config implementation."""

import enum
from collections.abc import Callable, Sequence
from dataclasses import field
from typing import Any

from pydantic.dataclasses import dataclass
from torch import nn

from rankfx.common.base.config.base_config import BaseConfig


@enum.unique
class ModelStructure(enum.Enum):
    """Possible model configurations."""
    CROSSNET_ONLY = "crossnet_only"
    STACKED = "stacked"
    PARALLEL = "parallel"
    STACKED_PARALLEL = "stacked_parallel"


@dataclass
class DCNv2Config(BaseConfig):
    """DCNv2 config implementation."""

    model_structure: ModelStructure = field(default=ModelStructure.STACKED)

    output_dim: int = field(default=1)

    loss_fn: str = field(default="torch.nn.BCEWithLogitsLoss")
    loss_params: dict[str, Any] = field(default_factory=lambda: {})

    # CrossNet parameters
    use_low_rank_mixture: bool = field(default=True)
    cross_low_rank_dim: int = field(default=32)
    num_cross_layers: int = field(default=4)
    num_cross_experts: int = field(default=4)

    # Parallel DNN parameters
    parallel_hidden_dims: Sequence[int] | None = field(default=None)
    parallel_dropout: float = field(default=0.)
    parallel_use_batch_norm: bool = field(default=True)
    parallel_activation: Callable = field(default=nn.ReLU)

    # Stacked DNN parameters
    stacked_hidden_dims: Sequence[int] | None = field(default=None)
    stacked_dropout: float = field(default=0.)
    stacked_use_batch_norm: bool = field(default=True)
    stacked_activation: Callable = field(default=nn.ReLU)

    def get_backbone_output_dim(self, input_dim: int | None = None) -> int:
        """Get output dim after CrossNet + DNN steps."""
        if self.model_structure == ModelStructure.CROSSNET_ONLY:
            return input_dim
        if self.model_structure == ModelStructure.STACKED:
            return self.stacked_hidden_dims[-1]
        if self.model_structure == ModelStructure.PARALLEL:
            return input_dim + self.parallel_hidden_dims[-1]
        return self.stacked_hidden_dims[-1] + self.parallel_hidden_dims[-1]



"""Module with FinalNet model class implementation."""

from collections.abc import Callable, Sequence
from typing import Any, Literal

import torch
from sklearn.metrics import log_loss, roc_auc_score
from torch import nn

from rankfx.common.base.model import NNPandasModel
from rankfx.common.features.config import FeaturesConfig
from rankfx.common.modules import EmbeddingLayer, FieldGate, FinalBlock
from rankfx.common.utils.import_utils import import_module_by_path


class FinalNet(NNPandasModel):
    """FinalNet implementation."""
    def __init__(
        self,
        loss_fn: str = "torch.nn.BCEWithLogitsLoss",
        loss_params: dict[str, Any] | None = None,
        block_type: Literal["1B", "2B"] = "2B",
        use_field_gate: bool = True,
        use_batch_norm: bool = True,
        add_bias: bool = True,
        block1_hidden_dims: Sequence[int] = (64, 64, 64),
        block1_hidden_activations: Sequence[Callable] | Callable | None = None,
        block1_dropout_rates: Sequence[float] | float | None = None,
        block2_hidden_dims: Sequence[int] = (64, 64, 64),
        block2_hidden_activations: Sequence[Callable] | Callable | None = None,
        block2_dropout_rates: Sequence[float] | float | None = None,
        residual_type: Literal["sum", "concat"] = "concat",
        proj_output_embeddings: bool = False,
        features_config: FeaturesConfig | None = None,
        oov_idx: int = 0,
    ):
        """Instantiate FinalNet module.

        Features are taken in the order, specified in features_config.features

        Args:
            loss_fn: loss function import path
            loss_params: model loss parameters
            block_type: type of blocks for FinalNet model
            use_field_gate: use FieldGate layer for the first block
            use_batch_norm: use batch norm for FinalBlock hidden layers
            add_bias: add bias to FactorizedInteraction layer
            block1_hidden_dims: hidden layers dimensions for block 1
            block1_hidden_activations: hidden layers activations for block 1
            block1_dropout_rates: hidden layers dropout rates for block 1
            block2_hidden_dims: hidden layers dimensions for block 2
            block2_hidden_activations: hidden layers activations for block 2
            block2_dropout_rates: hidden layers dropout rates for block 2
            residual_type: type of residual for FactorizedInteraction, one of `sum` and `concat`
            proj_output_embeddings: apply linear layer to concatted embeddings
            features_config: features config if known beforehand
            oov_idx: index to use as OOV for all categorical features embeddings
        """
        super().__init__(
            infer_feature_config=features_config is None,
            oov_idx=oov_idx,
        )

        if block_type not in ["1B", "2B"]:
            raise RuntimeError(f"block_type should be on of [`1B`, `2B`], got `{block_type}`")

        self.use_field_gate = use_field_gate
        self.use_batch_norm = use_batch_norm
        self.add_bias = add_bias
        self.block_type = block_type

        self.block1_hidden_dims = block1_hidden_dims
        self.block1_hidden_activations = block1_hidden_activations
        self.block1_dropout_rates = block1_dropout_rates

        self.block2_hidden_dims = block2_hidden_dims
        self.block2_hidden_activations = block2_hidden_activations
        self.block2_dropout_rates = block2_dropout_rates

        self.residual_type = residual_type

        self.proj_output_embeddings = proj_output_embeddings

        loss_cls = import_module_by_path(loss_fn)
        self.loss_fn = loss_cls(**(loss_params or {}))

        if features_config is not None:
            self._init_modules(features_config=features_config)

    def _init_modules(self, features_config: FeaturesConfig) -> None:
        """Instantiate FinalNet modules based on features config.

        Args:
            features_config: features config
        """
        self.embedding_layer = EmbeddingLayer(
            features_config=features_config,
            is_dict_input=True,
            proj_output_features=self.proj_output_embeddings,
            flatten_emb=False,
        )

        input_dim = features_config.num_final_features
        num_fields = len(features_config.features)

        self.fields_gate = nn.Identity()
        if self.use_field_gate:
            self.fields_gate = FieldGate(num_fields=num_fields)

        self.block1 = FinalBlock(
            input_dim=input_dim if not self.use_field_gate else input_dim * 2,
            hidden_dims=self.block1_hidden_dims,
            hidden_activations=self.block1_hidden_activations,
            dropout_rates=self.block1_dropout_rates,
            use_batch_norm=self.use_batch_norm,
            add_bias=self.add_bias,
            residual_type=self.residual_type,
        )
        self.proj1 = nn.Linear(self.block1_hidden_dims[-1], 1)

        self.block2, self.proj2 = None, None
        if self.block_type == "2B":
            self.block2 = FinalBlock(
                input_dim=input_dim,
                hidden_dims=self.block2_hidden_dims,
                hidden_activations=self.block2_hidden_activations,
                dropout_rates=self.block2_dropout_rates,
                use_batch_norm=self.use_batch_norm,
                add_bias=self.add_bias,
                residual_type=self.residual_type,
            )
            self.proj2 = nn.Linear(self.block1_hidden_dims[-1], 1)

    def forward(self, input_: dict[str, torch.Tensor] | torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Args:
            input_: input tensor [{feature_name: feature_tensor} or tensor (batch_size, concat_features_len)]

        Returns:
            Tensor after FinalNet applying
        """
        embeddings_out = self.embedding_layer(input_)
        fields_gate_out = self.fields_gate(embeddings_out)
        block1_out = self.proj1(self.block1(fields_gate_out.flatten(start_dim=1)))

        if self.block_type == "1B":
            return block1_out

        block2_out = self.proj2(self.block2(embeddings_out.flatten(start_dim=1)))
        return (block1_out + block2_out) * 0.5

    def train_step(self, batch: dict[str, torch.Tensor] | torch.Tensor) -> dict[str, Any]:
        """Model training step.

        Args:
            batch: batch of data

        Returns:
            Loss/metrics after training step
        """
        target = batch["target"]
        logits = self.forward(batch).squeeze()
        loss = self.loss_fn(logits, target.float())

        return {"loss": loss}

    def val_step(self, batch: dict[str, torch.Tensor] | torch.Tensor) -> dict[str, Any]:
        """Model validation step.

        Args:
            batch: batch of data

        Returns:
            Loss/metrics after validation step
        """
        target = batch["target"]
        logits = self.forward(batch).squeeze()
        probs = torch.sigmoid(logits).detach().cpu().numpy()
        target = target.detach().cpu().numpy()

        roc_auc = roc_auc_score(target, probs)
        logloss = log_loss(target, probs)

        return {"AUC": roc_auc, "log_loss": logloss}

    def test_step(self, batch: dict[str, torch.Tensor] | torch.Tensor) -> dict[str, Any]:
        """Model testing step.

        Args:
            batch: batch of data

        Returns:
            Loss/metrics after testing step
        """
        target = batch["target"]
        logits = self.forward(batch).squeeze()
        probs = torch.sigmoid(logits).detach().cpu().numpy()
        target = target.detach().cpu().numpy()

        roc_auc = roc_auc_score(target, probs)
        logloss = log_loss(target, probs)

        return {"AUC": roc_auc, "log_loss": logloss}

    def inference_step(self, batch: dict[str, torch.Tensor] | torch.Tensor) -> Any:
        """Model inference step.

        Args:
            batch: batch of data

        Returns:
            Model inference output
        """
        logits = self.forward(batch).squeeze()
        return logits



"""Module with DCNv2 model class implementation."""

from collections.abc import Callable, Sequence
from typing import Any, Literal

import torch
from sklearn.metrics import log_loss, roc_auc_score
from torch import nn

from rankfx.common.base.model import NNPandasModel
from rankfx.common.features.config import FeaturesConfig
from rankfx.common.modules import CrossNetMix, CrossNetV2, EmbeddingLayer, MLPBlock
from rankfx.common.utils.import_utils import import_module_by_path
from rankfx.dcnv2.config.model_config import ModelStructure


class DCNv2(NNPandasModel):
    """DCNv2 implementation."""
    def __init__(
        self,
        loss_fn: str = "torch.nn.BCEWithLogitsLoss",
        loss_params: dict[str, Any] | None = None,
        model_structure: Literal["crossnet_only", "stacked", "parallel", "stacked_parallel"] = "stacked",
        use_low_rank_mixture: bool = True,
        cross_low_rank_dim: int = 32,
        num_cross_layers: int = 4,
        num_cross_experts: int = 4,
        parallel_hidden_dims: Sequence[int] | None = None,
        parallel_dropout: float = 0.,
        parallel_use_batch_norm: bool = True,
        parallel_activation: Callable = nn.ReLU,
        stacked_hidden_dims: Sequence[int] | None = None,
        stacked_dropout: float = 0.,
        stacked_use_batch_norm: bool = True,
        stacked_activation: Callable = nn.ReLU,
        output_dim: int = 1,
        proj_output_embeddings: bool = False,
        features_config: FeaturesConfig | None = None,
        oov_idx: int = 0,
    ):
        """Instantiate DCNv2 module.

        Features are taken in the order, specified in features_config.features

        Args:
            loss_fn: loss function import path
            loss_params: model loss parameters
            model_structure: model structure type
            use_low_rank_mixture: use low rank projections inside the net
            cross_low_rank_dim: size of low rank cross net projections
            num_cross_layers: number of cross layers
            num_cross_experts: number of cross experts
            parallel_hidden_dims: hidden layers dimensions of parallel net
            parallel_dropout: dropout value of parallel net
            parallel_use_batch_norm: user batch norm in parallel net
            parallel_activation: activation function in parallel net
            stacked_hidden_dims: hidden layers dimensions of stacked net
            stacked_dropout: dropout value of stacked net
            stacked_use_batch_norm: user batch norm in stacked net
            stacked_activation: activation function in stacked net
            output_dim: net output dimensions
            proj_output_embeddings: apply linear layer to concatted embeddings
            features_config: features config if known beforehand
            oov_idx: index to use as OOV for all categorical features embeddings
        """
        super().__init__(
            infer_feature_config=features_config is None,
            oov_idx=oov_idx,
        )

        if model_structure not in [
            ModelStructure.CROSSNET_ONLY.value,
            ModelStructure.STACKED.value,
            ModelStructure.PARALLEL.value,
            ModelStructure.STACKED_PARALLEL.value,
        ]:
            raise RuntimeError(f"Unknown model structure `{model_structure}`")

        if (
            parallel_hidden_dims is None
            and model_structure in [ModelStructure.PARALLEL.value, ModelStructure.STACKED_PARALLEL.value]
        ):
            raise RuntimeError("No hidden sizes provided for Parallel DNN")

        if (
            stacked_hidden_dims is None
            and model_structure in [ModelStructure.STACKED.value, ModelStructure.STACKED_PARALLEL.value]
        ):
            raise RuntimeError("No hidden sizes provided for Stacked DNN")

        self.model_structure = model_structure
        self.proj_output_embeddings = proj_output_embeddings

        self.use_low_rank_mixture = use_low_rank_mixture
        self.cross_low_rank_dim = cross_low_rank_dim
        self.num_cross_layers = num_cross_layers
        self.num_cross_experts = num_cross_experts

        self.parallel_hidden_dims = parallel_hidden_dims
        self.parallel_dropout = parallel_dropout
        self.parallel_use_batch_norm = parallel_use_batch_norm
        self.parallel_activation = parallel_activation

        self.stacked_hidden_dims = stacked_hidden_dims
        self.stacked_dropout = stacked_dropout
        self.stacked_use_batch_norm = stacked_use_batch_norm
        self.stacked_activation = stacked_activation

        self.output_dim = output_dim

        loss_cls = import_module_by_path(loss_fn)
        self.loss_fn = loss_cls(**(loss_params or {}))

        if features_config is not None:
            self._init_modules(features_config=features_config)

    def _init_modules(self, features_config: FeaturesConfig) -> None:
        """Instantiate DCNv2 modules based on features config.

        Args:
            features_config: features config
        """
        self.embedding_layer = EmbeddingLayer(
            features_config=features_config,
            is_dict_input=True,
            proj_output_features=self.proj_output_embeddings
        )
        input_dim = features_config.num_final_features

        if self.use_low_rank_mixture:
            self.crossnet = CrossNetMix(
                input_dim,
                low_rank_dim=self.cross_low_rank_dim,
                num_layers=self.num_cross_layers,
                num_experts=self.num_cross_experts,
            )
        else:
            self.crossnet = CrossNetV2(input_dim, num_layers=self.num_cross_layers)

        if self.model_structure in (ModelStructure.STACKED.value, ModelStructure.STACKED_PARALLEL.value):
            self.stacked_dnn = MLPBlock(
                in_features=input_dim,
                hidden_dims=self.stacked_hidden_dims,
                activation_fn=self.stacked_activation,
                dropout=self.stacked_dropout,
                use_batch_norm=self.stacked_use_batch_norm,
            )

        if self.model_structure in (ModelStructure.PARALLEL.value, ModelStructure.STACKED_PARALLEL.value):
            self.parallel_dnn = MLPBlock(
                in_features=input_dim,
                hidden_dims=self.parallel_hidden_dims,
                activation_fn=self.parallel_activation,
                dropout=self.parallel_dropout,
                use_batch_norm=self.parallel_use_batch_norm,
            )

        final_dim = self._get_backbone_output_dim(
            input_dim=input_dim,
            parallel_hidden_dims=self.parallel_hidden_dims,
            stacked_hidden_dims=self.stacked_hidden_dims,
        )
        self.fc = nn.Linear(final_dim, self.output_dim)

    def _get_backbone_output_dim(
        self,
        input_dim: int | None = None,
        parallel_hidden_dims: Sequence[int] | None = None,
        stacked_hidden_dims: Sequence[int] | None = None,
    ) -> int:
        """Get output dim after CrossNet + DNN steps.

        Args:
            input_dim: number of input dimensions
            model_structure: model structure type
            parallel_hidden_dims: hidden layers dimensions of parallel net
            stacked_hidden_dims: hidden layers dimensions of stacked net

        Returns:
            Number of output dimensions after crossnet and dnn
        """
        if self.model_structure == ModelStructure.CROSSNET_ONLY.value:
            return input_dim
        if self.model_structure == ModelStructure.STACKED.value:
            return stacked_hidden_dims[-1]
        if self.model_structure == ModelStructure.PARALLEL.value:
            return input_dim + parallel_hidden_dims[-1]
        return stacked_hidden_dims[-1] + parallel_hidden_dims[-1]

    def forward(self, input_: dict[str, torch.Tensor] | torch.Tensor) -> torch.Tensor:
        """Forward pass.

        Args:
            input_: input tensor [{feature_name: feature_tensor} or tensor (batch_size, concat_features_len)]

        Returns:
            Tensor after DCNv2 applying
        """
        embeddings_out = self.embedding_layer(input_)
        cross_out = self.crossnet(embeddings_out)
        if self.model_structure == ModelStructure.CROSSNET_ONLY.value:
            final_out = cross_out
        elif self.model_structure == ModelStructure.STACKED.value:
            final_out = self.stacked_dnn(cross_out)
        elif self.model_structure == ModelStructure.PARALLEL.value:
            final_out = torch.cat([cross_out, self.parallel_dnn(embeddings_out)], dim=-1)
        elif self.model_structure == ModelStructure.STACKED_PARALLEL.value:
            final_out = torch.cat([self.stacked_dnn(cross_out), self.parallel_dnn(embeddings_out)], dim=-1)
        else:
            raise RuntimeError(f"Unknown model structure: `{self.model_structure}`")

        return self.fc(final_out)

    def train_step(self, batch: dict[str, torch.Tensor] | torch.Tensor) -> dict[str, Any]:
        """Model training step.

        Args:
            batch: batch of data

        Returns:
            Loss/metrics after training step
        """
        target = batch["target"]
        logits = self.forward(batch).squeeze()
        loss = self.loss_fn(logits, target.float())

        return {"loss": loss}

    def val_step(self, batch: dict[str, torch.Tensor] | torch.Tensor) -> dict[str, Any]:
        """Model validation step.

        Args:
            batch: batch of data

        Returns:
            Loss/metrics after validation step
        """
        target = batch["target"]
        logits = self.forward(batch).squeeze()
        probs = torch.sigmoid(logits).detach().cpu().numpy()
        target = target.detach().cpu().numpy()

        roc_auc = roc_auc_score(target, probs)
        logloss = log_loss(target, probs)

        return {"AUC": roc_auc, "log_loss": logloss}

    def test_step(self, batch: dict[str, torch.Tensor] | torch.Tensor) -> dict[str, Any]:
        """Model testing step.

        Args:
            batch: batch of data

        Returns:
            Loss/metrics after testing step
        """
        target = batch["target"]
        logits = self.forward(batch).squeeze()
        probs = torch.sigmoid(logits).detach().cpu().numpy()
        target = target.detach().cpu().numpy()

        roc_auc = roc_auc_score(target, probs)
        logloss = log_loss(target, probs)

        return {"AUC": roc_auc, "log_loss": logloss}

    def inference_step(self, batch: dict[str, torch.Tensor] | torch.Tensor) -> Any:
        """Model inference step.

        Args:
            batch: batch of data

        Returns:
            Model inference output
        """
        logits = self.forward(batch).squeeze()
        return logits



"""Module with features configs implementation."""

from __future__ import annotations

from collections.abc import Sequence
from dataclasses import field

from pydantic import model_validator
from pydantic.dataclasses import dataclass

from rankfx.common.base.config.base_config import BaseConfig
from rankfx.common.features.types import FeatureType


@dataclass
class Feature(BaseConfig):
    """Config for single feature."""
    name: str
    feature_type: FeatureType
    feature_size: int = field(default=1)

    # embedding parameters
    needs_embed: bool = field(default=False)
    embedding_size: int | None = field(default=None)
    embedding_vocab_size: int | None = field(default=None)
    embedding_padding_idx: int | None = field(default=None)

    @model_validator(mode="after")
    def check_feature_size(self) -> Feature:
        """Validate feature size."""
        if (
            self.feature_size > 1
            and self.feature_type not in [FeatureType.NUMERICAL_SEQUENCE, FeatureType.CATEGORICAL_SEQUENCE]
        ):
            raise ValueError(f"Feature size cannot be > 1 for scalar feature types, got {self.feature_size}")
        return self


@dataclass
class FeaturesConfig(BaseConfig):
    """Config for model features."""
    features: Sequence[Feature] | None = field(default=None)

    @property
    def num_initial_features(self) -> int:
        """Get final number of features."""
        size_ = 0
        for feature in self.features:
            size_ += feature.feature_size

        return size_

    @property
    def num_final_features(self) -> int:
        """Get final number of features."""
        size_ = 0
        for feature in self.features:
            feature_size = feature.embedding_size if feature.needs_embed else feature.feature_size
            size_ += feature_size

        return size_
